import 'dart:async';
import 'dart:convert';

import 'package:analyzer/dart/analysis/utilities.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:build/build.dart';
import 'package:build_modules/build_modules.dart';
import 'package:build_node_compilers/build_node_compilers.dart';
import 'package:node_preamble/preamble.dart';
import 'package:path/path.dart' as p;
import 'package:scratch_space/scratch_space.dart';

const _jsEntrypointExtension = '.dart.js';
const _jsEntrypointSourceMapExtension = '.dart.js.map';

/// Compiles an the primary input of [buildStep] with dart2js.
///
/// If [skipPlatformCheck] is `true` then all `dart:` imports will be
/// allowed in all packages.
Future<void> bootstrapDart2Js(
  BuildStep buildStep,
  List<String> dart2JsArgs,
  Map nodeModules, {
  bool skipPlatformCheck = false,
}) async {
  var dartEntrypointId = buildStep.inputId;
  var moduleId =
      dartEntrypointId.changeExtension(moduleExtension(dart2jsPlatform));
  var args = <String>[];
  {
    var module = Module.fromJson(
        json.decode(await buildStep.readAsString(moduleId))
            as Map<String, dynamic>);
    List<Module> allDeps;
    try {
      allDeps = (await module.computeTransitiveDependencies(buildStep,
          throwIfUnsupported: !skipPlatformCheck))
        ..add(module);
    } on UnsupportedModules catch (e) {
      var librariesString = (await e.exactLibraries(buildStep).toList())
          .map((lib) => AssetId(lib.id.package,
              lib.id.path.replaceFirst(moduleLibraryExtension, '.dart')))
          .join('\n');
      log.warning('''
Skipping compiling ${buildStep.inputId} with dart2js because some of its
transitive libraries have sdk dependencies that not supported on this platform:

$librariesString

https://github.com/dart-lang/build/blob/master/docs/faq.md#how-can-i-resolve-skipped-compiling-warnings
''');
      return;
    }

    var scratchSpace = await buildStep.fetchResource(scratchSpaceResource);
    var allSrcs = allDeps.expand((module) => module.sources);
    await scratchSpace.ensureAssets(allSrcs, buildStep);

    var dartPath = dartEntrypointId.path.startsWith('lib/')
        ? 'package:${dartEntrypointId.package}/'
            '${dartEntrypointId.path.substring('lib/'.length)}'
        : dartEntrypointId.path;
    var jsOutputPath =
        '${p.withoutExtension(dartPath.replaceFirst('package:', 'packages/'))}'
        '$_jsEntrypointExtension';
    args = dart2JsArgs.toList()
      ..addAll([
        '--packages=${p.join('.dart_tool', 'package_config.json')}',
        '-o$jsOutputPath',
        dartPath,
      ]);
  }

  var dart2js = await buildStep.fetchResource(dart2JsWorkerResource);
  var result = await dart2js.compile(args);
  var jsOutputId = dartEntrypointId.changeExtension(_jsEntrypointExtension);
  var jsOutputFile = scratchSpace.fileFor(jsOutputId);
  if (result.succeeded && await jsOutputFile.exists()) {
    log.info(result.output);

    final buffer = StringBuffer();
    String preamble;
    if (nodeModules != null) {
      nodeModules.forEach((key, value) {
        buffer.write('const $key = require("$value");');
      });
      final globals = nodeModules.keys.toList().cast<String>();
      preamble = getPreamble(minified: true, additionalGlobals: globals);
    } else {
      preamble = getPreamble(minified: true);
    }
    final contents = jsOutputFile.readAsStringSync();
    buffer
      ..write(preamble)
      ..writeln(contents);
    jsOutputFile.writeAsStringSync(buffer.toString());

    await scratchSpace.copyOutput(jsOutputId, buildStep);
    var jsSourceMapId =
        dartEntrypointId.changeExtension(_jsEntrypointSourceMapExtension);
    await _copyIfExists(jsSourceMapId, scratchSpace, buildStep);
  } else {
    log.severe(result.output);
  }
}

Builder nodeJsBuilder(BuilderOptions options) =>
    NodeJsBuilder.fromOptions(options);

Future<void> _copyIfExists(
  AssetId id,
  ScratchSpace scratchSpace,
  AssetWriter writer,
) async {
  var file = scratchSpace.fileFor(id);
  if (await file.exists()) await scratchSpace.copyOutput(id, writer);
}

/// Returns whether or not [dartId] is an app entrypoint (basically, whether
/// or not it has a `main` function).
Future<bool> _isAppEntryPoint(AssetId dartId, AssetReader reader) async {
  assert(dartId.extension == '.dart');
  // Skip reporting errors here, dartdevc will report them later with nicer
  // formatting.
  // ignore: deprecated_member_use
  final parsed = parseString(
    content: await reader.readAsString(dartId),
    throwIfDiagnostics: false,
  );
  // Allow two or fewer arguments so that entrypoints intended for use with
  // [spawnUri] get counted.
  //
  return parsed.unit.declarations.any((node) =>
      node is FunctionDeclaration &&
      node.name.name == 'main' &&
      node.functionExpression.parameters.parameters.length <= 2);
}

class NodeJsBuilder extends NodeEntrypointBuilder {
  final Map nodeModules;

  const NodeJsBuilder(
    WebCompiler webCompiler, [
    List<String> dart2JsArgs,
    this.nodeModules,
  ]) : super(webCompiler, dart2JsArgs: dart2JsArgs);

  factory NodeJsBuilder.fromOptions(BuilderOptions options) {
    final dart2JsConfig = options.config['dart2js_args'] as List;
    final dart2JsArgs = dart2JsConfig?.cast<String>() ?? <String>[];
    final nodeModules = options.config['node_modules'] as Map;
    return NodeJsBuilder(WebCompiler.Dart2Js, dart2JsArgs, nodeModules);
  }

  @override
  Future<void> build(BuildStep buildStep) async {
    var dartEntrypointId = buildStep.inputId;
    var isAppEntrypoint = await _isAppEntryPoint(dartEntrypointId, buildStep);
    if (!isAppEntrypoint) return;
    await bootstrapDart2Js(
      buildStep,
      dart2JsArgs,
      nodeModules,
      skipPlatformCheck: true,
    );
  }
}
